function watchEx(fn) {
    function wrapper() {
        try {
            return fn.apply(this, arguments);
        } catch (ex) {
            window.YoutubeSeek && window.YoutubeSeek.recordError();
            throw ex;
        }
    }

    return augmentFunctionName(wrapper, fn, 'watchEx.');
}

/**
 * This is a hack that attempts to "change" a function's "name" property. The name gets used in stack traces, so a name that's better than "anonymous function" is always helpful.
 * This hack is useful when wrapping other functions, because you can take the name of the wrapped function and copy it over to the wrapper you created, allowing you to kinda maintain the name.
 *
 * @example
 * function updateGui() {
 *     // ...
 * }
 * console.log(updateGui.name); // "updateGui"
 * let wrapped = debounce(updateGui);
 * console.log(wrapped.name); // "anonymous"
 * augmentFunctionName(wrapped, updateGui, 'debouncedFunc.');
 * console.log(wrapped.name); // "debouncedFunc.updateGui"
 *
 *
 * @param {function} fnToReceiveNewName
 * @param {function} fnToPartiallyCopyNameFrom
 * @param {string?} fnNamePrefix
 */
function augmentFunctionName(fnToReceiveNewName, fnToPartiallyCopyNameFrom, fnNamePrefix = 'wrapper.') {
    // Hack to set the function name of the wrapper so that it contains the name of the wrapped function - for better stack traces.
    Object.defineProperty(fnToReceiveNewName, 'name', {value: fnNamePrefix + fnToPartiallyCopyNameFrom.name});

    // We don't need to return the object because we modified the existing object, but we do it to make using/calling this function easier.
    return fnToReceiveNewName;
}

function debounce(func, delayUntilCallMillis) {
    let timerId;
    function wrapper() {
        const that = this;
        const args = arguments;
        timerId && clearTimeout(timerId);
        timerId = setTimeout(function debounceWrapper() {
            timerId = undefined;
            func.apply(that, args);
        }, delayUntilCallMillis);
    }
    return augmentFunctionName(wrapper, func, 'debounceWrapper.');
}

/**
 *
 from https://stackoverflow.com/questions/27078285/simple-throttle-in-js
 Returns a function, that, when invoked, will only be triggered at most once
 during a given window of time. Normally, the throttled function will run
 as much as it can, without ever going more than once per `wait` duration;
 but if you'd like to disable the execution on the leading edge, pass
 `{leading: false}`. To disable execution on the trailing edge, ditto.

 * @param {function(): *} func
 * @param {number} minimumMillisBetween
 * @param {object} options
 * @return {function(): *}
 */
function throttle(func, minimumMillisBetween, options) {
    let context, args, result;
    let timeout = null;
    let previous = 0;
    if (!options) options = {};
    let later = function() {
        previous = options.leading === false ? 0 : Date.now();
        timeout = null;
        result = func.apply(context, args);
        if (!timeout) context = args = null;
    };
    function throttledFunc() {
        let now = Date.now();
        if (!previous && options.leading === false) previous = now;
        let remaining = minimumMillisBetween - (now - previous);
        context = this;
        args = arguments;
        if (remaining <= 0 || remaining > minimumMillisBetween) {
            if (timeout) {
                clearTimeout(timeout);
                timeout = null;
            }
            previous = now;
            result = func.apply(context, args);
            if (!timeout) context = args = null;
        } else if (!timeout && options.trailing !== false) {
            timeout = setTimeout(watchEx(later), remaining);
        }
        return result;
    }
    return augmentFunctionName(throttledFunc, func, 'throttleWrapper.');
}

/**
 *
 * @param {function} func
 * @param minimumMillisBetween
 * @return {function}
 */
function leadingAndTrailingFiringThrottle(func, minimumMillisBetween) {
    return throttle(func, minimumMillisBetween, {leading: true, trailing: true});
}

function byId(domId) {
    return document.getElementById(domId);
}

/**
 * @param {string} selector
 * @param {Node?} contextNode
 * @return {Node[]}
 */
function queryAll(selector, contextNode) {
    return Array.from((contextNode || document).querySelectorAll(selector));
}

/**
 * @param {string} selector
 * @param {Node?} contextNode
 * @return {Node|null}
 */
function query(selector, contextNode) {
    return (contextNode || document).querySelector(selector);
}


export {debounce, throttle, leadingAndTrailingFiringThrottle, byId, watchEx, query, queryAll, augmentFunctionName};
